macroscript SelectByMeasuring
category:"BodyulCG"
tooltip:"Select By Measuring"
(	
    local dialogPos
    local dialogParams = #()
    
    rollout SelectByMeasuring "Select By Measuring" width:176
    (
        local M_BOUNDINGBOX = 1
        local M_SURFACEAREA = 2
        local M_VOLUME = 3
        local M_SHAPELENGTH = 4
        
        local COND_EQUAL = 1
        local COND_NOTEQUAL = 2
        local COND_LESS = 3
        local COND_LESSEQUAL = 4
        local COND_GREATEREQUAL = 5
        local COND_GREATER = 6
        
        group "Select By"
        (
            dropdownlist ddlSelectBy items:#("Bounding Box", "Surface Area", "Volume", "Shape Length") 
        )
        
        group "Condition"
        (
            dropdownlist ddlCondition items:#("Equal To", "Not Equal To", "Less Than", "Less Than or Equal To", "Greater Than or Equal To", "Greater Than")
        )
        
        group "Parameters"
        (
            button btnFill "Fill From Selection" width:156 height:22 align:#center
            spinner spnLength "Length:" range:[0, 1e7, 0]
            spinner spnWidth "Width:" range:[0, 1e7, 0]
            spinner spnHeight "Height:" range:[0, 1e7, 0]
            spinner spnSurfaceArea "Surface Area:" range:[0, 1e7, 0]
            spinner spnVolume "Volume:" range:[0, 1e7, 0]
            spinner spnShapeLength "Shape Length:" range:[0, 1e7, 0]
            spinner spnTolerance "Tolerance:" range:[0, 1e7, 0.001]
        )
        
        group "Object Category"
        (
            checkbox chkCatGeometry "Geometry" checked:true
            checkbox chkCatShapes "Shapes" checked:true
        )
        
        group "Select"
        (
            checkbox chkVisibleObjectsOnly "Visible Objects Only" checked:true
            button btnSelect "Select" width:156 height:28 offset:[0, 4] align:#center tooltip:"Hold CTRL - Add to Selection\nHold ALT - Remove from Selection"
        )
        
        group "About"
        (
            label lblName "Select By Measuring v1.2"
            hyperlink hpUrl "bodyulcg.com" \
                address:"https://bodyulcg.com" \
                color:(color 0 240 150) \
                hoverColor:(color 0 240 150) \
                visitedColor:(color 0 240 150) \
                align:#center
        )

        fn updateUI =
        (
            spnWidth.enabled = ddlSelectBy.selection == M_BOUNDINGBOX
            spnLength.enabled = ddlSelectBy.selection == M_BOUNDINGBOX
            spnHeight.enabled = ddlSelectBy.selection == M_BOUNDINGBOX
            spnSurfaceArea.enabled = ddlSelectBy.selection == M_SURFACEAREA
            spnVolume.enabled = ddlSelectBy.selection == M_VOLUME
            spnShapeLength.enabled = ddlSelectBy.selection == M_SHAPELENGTH
            spnTolerance.enabled = finditem #(COND_EQUAL, COND_NOTEQUAL, COND_LESSEQUAL, COND_GREATEREQUAL) ddlCondition.selection > 0
            
            chkCatGeometry.enabled = ddlSelectBy.selection != M_SHAPELENGTH
            chkCatShapes.enabled = finditem #(M_BOUNDINGBOX, M_SHAPELENGTH) ddlSelectBy.selection > 0
        )
        
        fn calcOBBSize node =
        (
            local corners = nodeGetBoundingBox node node.transform
            local width = corners[2].x - corners[1].x
            local length = corners[2].y - corners[1].y
            local height = corners[2].z - corners[1].z
            
            return [width, length, height]
        )
                
        fn calcSurfaceArea node =
            return meshop.getFaceArea node.mesh #{1..node.mesh.numfaces};
        
        fn calcVolume node =
        (
            local volume = 0.0
            local theMesh = if isKindOf node Editable_Mesh then node.mesh else copy node.mesh;
            local numFaces = theMesh.numfaces
            
            for i = 1 to numFaces do
            (
                local face = getFace theMesh i
                local vert2 = getVert theMesh face.z
                local vert1 = getVert theMesh face.y
                local vert0 = getVert theMesh face.x
                
                volume += dot (cross (vert1 - vert0) (vert2 - vert0)) vert0
            )
            
            if not isKindOf node Editable_Mesh then
                free theMesh;

            return volume /= 6.0
        )
        
        fn matchOBB node width length height condition:COND_EQUAL tolerance:0 =
        (
            local obbSize = calcOBBSize node
            
            return case condition of
            (
                COND_EQUAL: (
                    abs (width - obbSize.x) <= tolerance and \
                    abs (length - obbSize.y) <= tolerance and \
                    abs (height - obbSize.z) <= tolerance
                )
                COND_NOTEQUAL: (
                    abs (width - obbSize.x) >= tolerance and \
                    abs (length - obbSize.y) >= tolerance and \
                    abs (height - obbSize.z) >= tolerance
                )
                COND_LESS: obbSize.x < width and obbSize.y < length and obbSize.z < height
                COND_LESSEQUAL: (obbSize.x <= width + tolerance) and (obbSize.y <= length + tolerance) and (obbSize.z <= height + tolerance)
                COND_GREATEREQUAL: (obbSize.x >= width - tolerance) and (obbSize.y >= length - tolerance) and (obbSize.z >= height - tolerance)
                COND_GREATER: obbSize.x > width and obbSize.y > length and obbSize.z > height
                default:false
            )
        )
        
        fn matchSurfaceArea mesh area condition:COND_EQUAL tolerance:0 =
        (
            local surfArea = calcSurfaceArea mesh
            
            return case condition of
            (
                COND_EQUAL: abs (area - surfArea) <= tolerance
                COND_NOTEQUAL: abs (area - surfArea) >= tolerance
                COND_LESS: surfArea < area
                COND_LESSEQUAL: surfArea <= area + tolerance
                COND_GREATEREQUAL: surfArea >= area - tolerance
                COND_GREATER: surfArea > area
                default:false
            )
        )
        
        fn matchVolume mesh volume condition:COND_EQUAL tolerance:0 =
        (
            local meshVolume = calcVolume mesh

            return case condition of
            (
                COND_EQUAL: abs (volume - meshVolume) <= tolerance
                COND_NOTEQUAL: abs (volume - meshVolume) >= tolerance
                COND_LESS: meshVolume < volume
                COND_LESSEQUAL: meshVolume <= volume + tolerance
                COND_GREATEREQUAL: meshVolume >= volume - tolerance
                COND_GREATER: meshVolume > volume
                default:false
            )
        )
        
        fn matchShapeLength shape length condition:COND_EQUAL tolerance:0 =
        (
            local shapeLength = curveLength shape
            
            return case condition of
            (
                COND_EQUAL: abs (length - shapeLength) <= tolerance
                COND_NOTEQUAL: not abs (length - shapeLength) <= tolerance
                COND_LESS: shapeLength < length
                COND_LESSEQUAL: shapeLength < length + tolerance
                COND_GREATEREQUAL: shapeLength > length - tolerance
                COND_GREATER: shapeLength > length
                default:false
            )
        )

        on SelectByMeasuring open do
        (
            if dialogPos != undefined do
                setDialogPos SelectByMeasuring dialogPos;
            
            for i in dialogParams collect
                setProperty (getProperty SelectByMeasuring i[1]) i[2] i[3]
            
            updateUI()
        )
    
        on ddlSelectBy selected sel do
            updateUI();

        on ddlCondition selected sel do
            updateUI();
        
        on btnFill pressed do
        (
            if isValidNode $ then
            (
                case ddlSelectBy.selection of
                (
                    M_BOUNDINGBOX: (
                        if isKindOf $ Node then 
                        (
                            local obbSize = calcOBBSize $
                            spnWidth.value = obbSize.x
                            spnLength.value = obbSize.y
                            spnHeight.value = obbSize.z
                        )
                    )
                    M_SURFACEAREA: (
                        if isKindOf $ GeometryClass and canConvertTo $ Editable_Mesh then 
                            spnSurfaceArea.value = calcSurfaceArea $;
                    )
                    
                    M_VOLUME: (
                        if isKindOf $ GeometryClass and canConvertTo $ Editable_Mesh then 
                            spnVolume.value = calcVolume $;
                    )
                    M_SHAPELENGTH: (
                        if isKindOf $ Shape then 
                            spnShapeLength.value = curveLength $;
                    )
                )
            )
        )
            
        on btnSelect pressed do
        (
            local condition = ddlCondition.selection
            local tolerance = spnTolerance.value
            local visibleOnly = chkVisibleObjectsOnly.checked
            local matchingObjects = #()
            local objs = #()
            local canceled = false
            
            setWaitCursor()
            
            case ddlSelectBy.selection of
            (
                M_BOUNDINGBOX: (
                    local width = spnWidth.value
                    local length = spnLength.value
                    local height = spnHeight.value
                    
                    if chkCatGeometry.checked then
                        join objs geometry;
                    
                    if chkCatShapes.checked then
                        join objs shapes;
                        
                    for i in objs while not canceled where visibleOnly and not i.isHiddenInVpt or not visibleOnly do
                    (
                        if matchOBB i width length height condition:condition tolerance:tolerance do
                            append matchingObjects i;
                        
                        canceled = keyboard.escPressed
                    )
                )
                M_SURFACEAREA: (
                    local area = spnSurfaceArea.value

                    if chkCatGeometry.checked then
                        join objs geometry;
                        
                    for i in objs while not canceled where visibleOnly and not i.isHiddenInVpt or not visibleOnly and \
                        canConvertTo i Editable_Mesh do
                    (
                        if matchSurfaceArea i area condition:condition tolerance:tolerance do
                            append matchingObjects i;
                        
                        canceled = keyboard.escPressed
                    )
                )
                M_VOLUME: (
                    local volume = spnVolume.value

                    if chkCatGeometry.checked then join objs geometry;
                        
                    for i in objs while not canceled where visibleOnly and not i.isHiddenInVpt or not visibleOnly and \
                        canConvertTo i Editable_Mesh do
                    (                        
                        if matchVolume i volume condition:condition tolerance:tolerance do
                            append matchingObjects i;
                        
                        canceled = keyboard.escPressed
                    )
                )
                M_SHAPELENGTH: (
                    local length = spnShapeLength.value

                    if chkCatShapes.checked then join objs shapes;
                        
                    for i in objs while not canceled where visibleOnly and not i.isHiddenInVpt or not visibleOnly do
                    (
                        if matchShapeLength i length condition:condition tolerance:tolerance do
                            append matchingObjects i;
                        
                        canceled = keyboard.escPressed
                    )
                )
            )
            
            free objs
            
            if not canceled then
            (
                undo on
                (
                    case of
                    (
                        (keyboard.controlPressed): selectmore matchingObjects
                        (keyboard.altPressed): deselect matchingObjects
                        default: (clearSelection(); select matchingObjects)
                    )
                )
            )
            
            return ok
        )
        
        on SelectByMeasuring close do
        (
            dialogPos = getDialogPos SelectByMeasuring
            dialogParams = for i in SelectByMeasuring.controls collect
            (
                case (classof i) of
                (
                    SpinnerControl: #(i.name, #value, i.value)
                    ComboBoxControl: #(i.name, #selection, i.selection)
                    CheckBoxControl: #(i.name, #checked, i.checked)
                    default: dontcollect
                )
            )
            
            gc light:true
        )
    )
    
    on execute do
    (
        if SelectByMeasuring.open then
            destroyDialog SelectByMeasuring;
        else
            createDialog SelectByMeasuring;
    )
)